import { InjectionToken, ɵɵdefineInjectable, ɵɵinject, Injectable, Inject } from '@angular/core';
import { Router, ActivatedRoute } from '@angular/router';
import { Subject, identity, merge } from 'rxjs';
import { startWith, map, auditTime, takeUntil, pairwise } from 'rxjs/operators';
import set from 'lodash.set';

const BIND_QUERY_PARAMS_OPTIONS = new InjectionToken('BIND_QUERY_PARAMS_OPTIONS', {
    providedIn: 'root',
    factory() {
        return {
            windowRef: window,
        };
    },
});

function isEmptyValue(def, value) {
    switch (def.type) {
        case 'array':
            return !value || !value.length;
        case 'object':
            return !value || !Object.keys(value).length;
        case 'string':
            return !value || value === '';
        default:
            return false;
    }
}
function parse(value, type) {
    switch (type) {
        case 'string':
            return value;
        case 'boolean':
            return value === 'false' ? false : !!value;
        case 'array':
            return value === null || value === void 0 ? void 0 : value.split(',');
        case 'number':
            return +value;
        case 'object':
            return JSON.parse(value);
        default:
            return value;
    }
}
function get(obj, path) {
    let current = obj;
    path.split('.').forEach((p) => (current = current[p]));
    return current;
}
function resolveParams(params) {
    const toArray = coerceArray(params);
    const result = {};
    toArray.forEach(({ def, value }) => {
        let defValue;
        if (def.removeEmptyValue) {
            defValue = isEmptyValue(def, value) ? null : def.serialize(value);
        }
        else {
            defValue = def.serialize(value);
        }
        result[def.queryKey] = defValue;
    });
    return result;
}
function coerceArray(value) {
    return Array.isArray(value) ? value : [value];
}
function defsToParams(defs, group) {
    return defs.map((def) => {
        return {
            queryKey: def.queryKey,
            value: group.get(def.path).value,
        };
    });
}

class QueryParamDef {
    constructor(config) {
        this.config = config;
    }
    get queryKey() {
        return this.config.queryKey;
    }
    get path() {
        return this.config.path || this.queryKey;
    }
    get type() {
        return this.config.type || 'string';
    }
    get strategy() {
        return this.config.strategy || 'twoWay';
    }
    get parser() {
        return this.config.parser;
    }
    get serializer() {
        return this.config.serializer;
    }
    get syncInitialValue() {
        return this.config.syncInitialValue || false;
    }
    get removeEmptyValue() {
        return this.config.removeEmptyValue === undefined ? true : this.config.removeEmptyValue;
    }
    serialize(controlValue) {
        if (this.serializer) {
            return this.serializer(controlValue);
        }
        if (controlValue === null || controlValue === undefined || controlValue === '') {
            return null;
        }
        const serializedValue = controlValue.toString();
        return serializedValue === '[object Object]' ? JSON.stringify(controlValue) : serializedValue;
    }
    parse(queryParamValue) {
        if (this.parser) {
            return this.parser(queryParamValue);
        }
        return parse(queryParamValue, this.type);
    }
}

class BindQueryParamsManager {
    constructor(router, activeRoute, defs, options) {
        this.router = router;
        this.activeRoute = activeRoute;
        this.options = options;
        this.$destroy = new Subject();
        this.defsSynced = {};
        this.defs = coerceArray(defs).map((def) => new QueryParamDef(def));
    }
    connect(group) {
        this.group = group;
        this.onInit();
        return this;
    }
    onInit() {
        this.updateControl(this.defs, { emitEvent: true }, shouldSyncInitialValue);
        const controls = this.defs.map((def) => {
            const control = this.group.get(def.path);
            return control.valueChanges.pipe(shouldSyncInitialValue(def) ? startWith(control.value) : identity, map((value) => ({
                def,
                value,
            })));
        });
        // Could be a several changes in the same tick,
        // for example when we use reset() or patchValue.
        // We need to aggregate the changes and apply them at once
        // because the router navigates in micro task
        let buffer = [];
        merge(...controls)
            .pipe(map((result) => buffer.push(result)), auditTime(0), takeUntil(this.$destroy))
            .subscribe(() => {
            this.updateQueryParams(resolveParams(buffer));
            buffer = [];
        });
        const twoWaySyncDef = this.defs.filter(({ strategy }) => strategy === 'twoWay');
        if (twoWaySyncDef.length) {
            this.activeRoute.queryParams
                .pipe(pairwise(), takeUntil(this.$destroy))
                .subscribe(([prevQueryParams, curQueryParams]) => {
                let paramsDiff = twoWaySyncDef.filter(({ queryKey }) => prevQueryParams[queryKey] !== curQueryParams[queryKey]);
                if (paramsDiff.length) {
                    this.updateControl(paramsDiff, { emitEvent: true });
                }
            });
        }
    }
    destroy() {
        this.$destroy.next();
    }
    getDef(queryKey) {
        return this.defs.find((def) => def.queryKey === queryKey);
    }
    parse(queryParams) {
        const result = {};
        for (const [key, value] of Object.entries(queryParams)) {
            const def = this.getDef(key);
            if (def) {
                result[key] = def.parse(value);
            }
        }
        return result;
    }
    syncAllDefs(options = { emitEvent: true }) {
        const allKeys = this.defs.map((def) => def.queryKey);
        this.syncDefs(allKeys, options);
    }
    syncDefs(queryKeys, options = { emitEvent: true }) {
        const defs = [];
        coerceArray(queryKeys).forEach((key) => {
            if (!this.defsSynced[key]) {
                this.defsSynced[key] = true;
                const def = this.getDef(key);
                if (def) {
                    defs.push(def);
                }
            }
        });
        if (defs.length) {
            this.updateControl(defs, options);
        }
    }
    paramExists(queryKey) {
        return new URLSearchParams(this.options.windowRef.location.search).has(queryKey);
    }
    someParamExists() {
        return this.defs.some((def) => {
            return new URLSearchParams(this.options.windowRef.location.search).has(def.queryKey);
        });
    }
    updateQueryParams(queryParams) {
        this.router.navigate([], {
            queryParams,
            queryParamsHandling: 'merge',
            replaceUrl: true,
        });
    }
    updateControl(defs, options, updatePredicate = (_) => true) {
        const queryParams = new URLSearchParams(this.options.windowRef.location.search);
        let value = {};
        for (const def of defs) {
            if (updatePredicate(def)) {
                const { queryKey } = def;
                const queryDef = queryParams.get(queryKey);
                if (!queryDef)
                    continue;
                set(value, def.path.split('.'), def.parse(queryDef));
            }
        }
        if (Object.keys(value).length) {
            this.group.patchValue(value, options);
        }
    }
}
function shouldSyncInitialValue(def) {
    return def.strategy === 'twoWay' || def.syncInitialValue;
}

class BindQueryParamsFactory {
    constructor(router, activeRoute, options) {
        this.router = router;
        this.activeRoute = activeRoute;
        this.options = options;
    }
    create(defs) {
        return new BindQueryParamsManager(this.router, this.activeRoute, defs, this.options);
    }
}
BindQueryParamsFactory.ɵprov = ɵɵdefineInjectable({ factory: function BindQueryParamsFactory_Factory() { return new BindQueryParamsFactory(ɵɵinject(Router), ɵɵinject(ActivatedRoute), ɵɵinject(BIND_QUERY_PARAMS_OPTIONS)); }, token: BindQueryParamsFactory, providedIn: "root" });
BindQueryParamsFactory.decorators = [
    { type: Injectable, args: [{ providedIn: 'root' },] }
];
BindQueryParamsFactory.ctorParameters = () => [
    { type: Router },
    { type: ActivatedRoute },
    { type: undefined, decorators: [{ type: Inject, args: [BIND_QUERY_PARAMS_OPTIONS,] }] }
];

/**
 * Generated bundle index. Do not edit.
 */

export { BIND_QUERY_PARAMS_OPTIONS, BindQueryParamsFactory };
//# sourceMappingURL=ngneat-bind-query-params.js.map
